/* Michael J. Chappell  February 6, 1993  FREDKIN.C

   Demonstrates four-fold reproduction using a system described by Edward
   Fredkin of MIT.  If you add the number of cells above, below, to the
   left and to the right of a cell, and the sum is odd then the cell will
   be filled, otherwise, the cell will be empty.

   Version 2.0 using ncurses - December 8, 2021

   Version 2.1 - December 19, 2021
   - Added load, save, pause, help, verbose, modify speed

   Version 2.2 - December 21, 2021
   - Various bug fixes

   Version 3.0 - January 4, 2022
   - Process lines in data files by first character:
     C - set cell alive character (ASCII value),
     D - set delay in seconds (value is a double and fractions of a 
         second can be used.)
     E - set empty cell character (ASCII value),
     G - Set value of generation,
     # - discard line as comment,
     digits - read as data,
     0 0 - end of data.
   - Now uses nanosleep and delay is input in seconds.  
     You should comment out delays in files saved under version 2.x 
     or earlier, or replace the value with the number of seconds.
   - Delay should be in range (0.0, 86400.0) (1 day = 86,400 seconds.)
   - Keys to adjust timing adjusted to work with seconds or fractions of 
     a second.
   - If a file is not loaded it will open with the menu.
*/

#include <ncurses/ncurses.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define VERSION_NUMBER "3.0"
#define COPYRIGHT "Copyright 2022     Michael J. Chappell"
#define DEFAULT_DELAY 0.1
#define DEFAULT_CELL '@'
#define DEFAULT_NOCELL ' '

/*
  We want to set defaults for CELL and NOCELL; however, we want to
  allow them to be refined by the user with the menu or set from load
  file.
*/
char CELL   = DEFAULT_CELL;
char NOCELL = DEFAULT_NOCELL;

/* 
   DEFAULT_FILENAME could be customized for the OS.  Avoid path components 
   here to simplify Windows and *NIX compatability.
*/
#define DEFAULT_FILENAME ""

/* 
   MAX_PATH_LENGTH is being set for the buffer and should be larger
   than what the OS allows.  I'm trying to reserve memory here to avoid
   using malloc and prevent problems from excessively long command lines.
*/
#define MAX_PATH_LENGTH 2048 

int buffer[2][27][82];

/* 
   Initialize(), Load_File() and Save_File() should all return 0
   Any other result indicates a file I/O problem we should terminate.
*/
int Input(int *generation, double *delay);
int Load_File(int *generation, double *delay, char *datafile);
int Save_File(int generation, double delay, char *datafile);

void Help(int generation, double delay, int argc, char *datafile);

void Buffer_To_Screen(int s);
void Redraw_Buffer(int s, int d);

void waitDSeconds(double n);

int  main(int argc, char **argv) {
     int board=0, generation=1;
     int c;
     int tempint;
     double delay=DEFAULT_DELAY;
     char datafile[MAX_PATH_LENGTH];
     int verbose=FALSE;
     int done=FALSE;
     int data=FALSE;
     
     initscr();

     if (argc>1){
       strncpy(datafile, argv[1], MAX_PATH_LENGTH);
       if (Load_File(&generation, &delay, datafile)) {
	 done = TRUE; /* Load_file should return 0 if successful */
       } /* if Load_File */
       data=TRUE;
     } else {
       strncpy(datafile, DEFAULT_FILENAME, MAX_PATH_LENGTH);
       Help(generation, delay, argc, datafile);
     }
     
     noecho();

     do {
       if (data) { /* If we never loaded or input data don't waste cycles */
	 nodelay(stdscr,TRUE);
	 Buffer_To_Screen(board);
	 Redraw_Buffer(board, (board+1)&1);
	 board = (board+1)&1;

	 move(0,0);
	 if (verbose) {
	   printw("Generation: %7d Delay: %11.4lf File: %s",
		  generation, delay, datafile);
	 }
	 refresh();
	 generation++;
       } else {
	 nodelay(stdscr,FALSE);
	 move(0,0);
	 clear();
	 printw("Fredkin Version %s\n", VERSION_NUMBER);
	 printw("%s\n", COPYRIGHT);
	 printw("\n");
	 if (verbose) {
	   printw("Generation: %7d Delay: %7d",
		  generation, delay);
	 }
	 printw("\n");
	 printw(
		"NO DATA PRESENT!\n"
		"\n"
		"\t  H) Help.\n"
		"\t  I) Input.\n"
		"\t  L) Load.\n"
		"\t  Q) Quit.\n\n"
		);
	 refresh();
       }
       c = getch();
       switch (tolower(c)) {
       case '0':
	 delay = DEFAULT_DELAY;
	 break;
       case '8':
	 delay = 0.0078125;
	 break;
       case '7':
	 delay = 0.015625;
	 break;
       case '6':
	 delay = 0.03125;
	 break;
       case '5':
	 delay = 0.0625;
	 break;
       case '4':
	 delay = 0.125;
	 break;
       case '3':
	 delay = 0.25;
	 break;
       case '2':
	 delay = 0.5;
	 break;
       case '1':
	 delay = 1.0;
	 break;
       case 'c':
	 nodelay(stdscr,FALSE);
	 echo();
	 clear();
	 printw("Enter ASCII code for a live cell.  \n");
	 refresh();
	 scanw("%d", &tempint);
	 CELL=(char)tempint;
	 nodelay(stdscr,TRUE);
	 noecho();
	 break;
       case 'd':
	 nodelay(stdscr,FALSE);
	 echo();
	 clear();
	 printw("Enter new delay in seconds.  \n");
	 refresh();
	 scanw("%lf", &delay);
	 nodelay(stdscr,TRUE);
	 noecho();
	 break;
       case 'e':
	 nodelay(stdscr,FALSE);
	 echo();
	 clear();
	 printw("Enter ASCII code for an empty cell.  \n");
	 refresh();
	 scanw("%d", &tempint);
	 NOCELL=(char)tempint;
	 nodelay(stdscr,TRUE);
	 noecho();
	 break;
       case 'h':
       case '?':
       case '\027':
	 Help(generation, delay, argc, datafile);
	 break;
       case 'i':
	 strncpy(datafile, DEFAULT_FILENAME, MAX_PATH_LENGTH);
	 if (Input(&generation, &delay)) done=TRUE;
	 data=TRUE;
	 break;
       case 'l':
	 nodelay(stdscr,FALSE);
	 echo();
	 clear();
	 printw("Load data from:  \n");
	 refresh();
	 scanw("%s", datafile);
	 noecho();
	 nodelay(stdscr,TRUE);
	 if (Load_File(&generation, &delay, datafile)) done=TRUE;
	 data=TRUE;
	 break;
       case 'p':
       case ' ':
	 refresh();
	 echo();
	 nodelay(stdscr,FALSE);
	 getch();
	 refresh();
	 nodelay(stdscr,TRUE);
	 noecho();
	 break;
       case 's':
	 nodelay(stdscr,FALSE);
	 echo();
	 clear();
	 printw("Save data to:  \n");
	 refresh();
	 scanw("%s", datafile);
	 nodelay(stdscr,TRUE);
	 noecho();
	 if (Save_File(generation, delay, datafile)) done=TRUE;
	 break;
       case 'v':
	 verbose = !verbose;
	 break;
       case 'q':
	 done = TRUE;
	 break;
	   
       case '-':
	 delay -= 1.0;
	 break;
       case '+':
	 delay -= 1.0;
	 break;
       case '<':
	 delay=(delay>0.0)?(delay/2):0.0;
	 break;
       case '>':
	 delay *= 2;
	 break;
       }
       /* delay cannot be less than 0 */
       delay=(delay>0)?delay:0;
       /* arbitrarily limit delay to 1 day */
       delay=(delay>(60*60*24))?(60*60*24):delay;
	 
       waitDSeconds(delay);
     } while (!done);

     echo();
     nodelay(stdscr,FALSE);
     getch();
     endwin();

     return EXIT_SUCCESS;
} /* main */

int Input(int *generation, double *delay) {
     int i, j, k;

     *delay = DEFAULT_DELAY;
     *generation = 1;
     

     for (k=0; k<2; k++)
       for (j=0; j<27; j++)
	 for (i=0; i<82; i++) buffer[k][j][i]=0;
     nodelay(stdscr,FALSE);
     echo();
     clear();
     do {
       addstr("\nEnter 0 0 to end input.  Otherwise, the specified ");
       addstr("\ncoordinates will be set if in range. [1,80] [1,25]  ");
       refresh();
       scanw("%d %d", &i, &j);
       if ((0<i)&&(i<81)&&(0<j)&&(j<26)) buffer[0][j][i]=1;
     } while (i!=0);
     nodelay(stdscr,TRUE);
     noecho();
     return 0;
} /* Initialize */

int Load_File(int *generation, double *delay, char *datafile) {
  FILE *fp;
  char line[MAX_PATH_LENGTH];
  int i, j, k;
  int tempint;
  int done=FALSE;

  CELL=DEFAULT_CELL;
  NOCELL=DEFAULT_NOCELL;
  for (k=0; k<2; k++)
    for (j=0; j<27; j++)
      for (i=0; i<82; i++) buffer[k][j][i]=0;
  /* 
     Error checking needs to be implemented.
  */
  if (fp=fopen(datafile,"r")) {
    do {
      fscanf(fp, "%[^\n]\n", line);
      /* printw("%s\n",line); refresh(); */
      switch (tolower(line[0])) {
      case 'c':
	sscanf(line, "%*s %i", &tempint);
	CELL=(char)tempint;
	break;
      case 'd':
	sscanf(line, "%*s %lf", delay);
	break;
      case 'e':
	sscanf(line, "%*s %i", &tempint);
	NOCELL=(char)tempint;
	break;
      case 'g':
	sscanf(line, "%*s %d", generation);
	break;
      case '#':
	break;
      default:
	sscanf(line, "%d %d", &i, &j);
	if ((0<i)&&(i<81)&&(0<j)&&(j<26)) buffer[0][j][i]=1;
	if (i==0) done=TRUE;
	break;
      }
    } while (!done);
    fclose(fp);
  } else {
    fprintf(stderr, "Unable to read from file: %s\n", datafile);
    return -1;
  }
  return 0;
}
  
int Save_File(int generation, double delay, char *datafile) {
  FILE *fp;
  int col, row;
  if (fp=fopen(datafile, "w")) {
    fprintf(fp, "# FREDKIN INPUT DATA MJC VERSION %s\n", VERSION_NUMBER);
    fprintf(fp, "Generation: %d\n", generation);
    fprintf(fp, "Delay: %lf\n", delay);
    fprintf(fp, "Cell: %d\n", CELL);
    fprintf(fp, "Empty: %d\n", NOCELL);
    for (row=1; row<26; row++) {
      for (col=1; col<81; col++) {
	if (buffer[0][row][col]) {
	  fprintf(fp, "%d %d\n", col, row);
	}
      }
    }
    fprintf(fp, "0 0\n");
    fclose(fp);
    return 0;
  } else {
    fprintf(stderr, "Unable to write to file: %s\n", datafile);
    return -1;
  }
}

void Help(int generation, double  delay, int argc, char *datafile) {
  char directory[MAX_PATH_LENGTH];

  getcwd(directory, MAX_PATH_LENGTH-1);
  clear();
  printw("Fredkin Version %s\n", VERSION_NUMBER);
  printw("%s\n", COPYRIGHT);
  if (argc==2) {
    printw("File              : %s\n", datafile);
  } else {
    printw("\n");
  }
  printw("Current directory : %s\n", directory);
  printw("Generation        : %10d", generation);
  printw("\t  Cell set   : '%c' (%3d)\n", CELL, CELL);
  printw("Delay (seconds)   : %10.4lf", delay);
  printw("\t  Cell clear : '%c' (%3d)\n", NOCELL, NOCELL);
  printw("Keys:\n"
	 "\t  C) Set ASCII Cell."
	 "\t  L) Load.      \n"
	 "\t  D) Set Delay. "
	 "\t  P) Pause.\n"
	 "\t  E) Set ASCII Empty."
	 "\t  Q) Quit.\n"
	 "\t  H) Help.      "
	 "\t  S) Save.\n"
	 "\t  I) Input Data."
	 "\t  V) Toggle Verbose\n\n"
	 "Keys to set Delay in seconds:\n"
	 "\t  0) Default Delay (%6.4lf seconds)\n"
	 "\t  1)   1.0000\t\t  5)    0.0625\n"
	 "\t  2)   0.5000\t\t  6)    0.0313\n"
	 "\t  3)   0.2500\t\t  7)    0.0156\n"
	 "\t  4)   0.1250\t\t  8)    0.0078\n\n"
	 "\t-/+) Decrement/Increment Delay by 1 second.\n"
	 "\t</>) Halve/Double Delay.\n\n"
	 "\t< Press any key to continue. >\n",
	 DEFAULT_DELAY
	 );
  refresh();
  echo();
  nodelay(stdscr,FALSE);
  getch();
  refresh();
  nodelay(stdscr,TRUE);
  noecho();

  return;
}

void Redraw_Buffer(int s, int d) {
     int i, j, temp;

     for (i=1; i<26; i++)
       for (j=1; j<81; j++) {
           temp = buffer[s][i][j-1]+buffer[s][i][j+1]+buffer[s][i-1][j];
           temp += buffer[s][i+1][j];
           buffer[d][i][j] = temp&1;
       } /* for j,i */
} /* Redraw_Buffer */

void Buffer_To_Screen(int s) {
     int i, j;

     clear();
     refresh();
     for (j=1; j<25; j++) {
       for (i=1; i<80; i++) {
           if ((j<25)|(i<80)) addch(buffer[s][j][i]?CELL:NOCELL);
       }
       addstr("\n");
     } /* for i,j */
     refresh();
} /* Buffer_To_Screen */

void waitDSeconds(double n) {
  long sec, nsec;
  struct timespec delay, remaining;
  
  if (n>0.0) {
    delay.tv_sec=(long)n; /* truncates decimal portion */
    delay.tv_nsec=1000000000*modf(n,NULL);
    nanosleep(&delay, &remaining);
  }
  return;
}
